#undef DEBUG_DMA

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "dmavif1.h"
#include "dmagif.h"
#include "screen.h" 
#include <geommath/geommath.h>
#include "camera.h"
#include "p2lassert.h"


int loop_again=1;

// this is executed when ctrl-c is pressed, ideally it stops the loop.
// the deinit functions have to be at the end of the main after the loop.
void sig_handle(int sig)
{
	loop_again=0;
}


#include <intmd/intmdloader.h>
#include "c3dsmesh.h"
#include "imesh.h"
#include "imeshinstance.h"

// global stuff
#define NUM_INSTANCES	5

CScreen *cScreen=NULL;
IMesh	myMesh;
IMeshInstance * pMeshInst[NUM_INSTANCES];

void SyncStitch(CDmaWrap * pDmaWrap)
{
	uint32 uCur = (uint32) pDmaWrap->GetCurLocation();
	assert((uCur&0xf)==0);
	if((uCur&0xff0)==0xff0)
	{
		pDmaWrap->AttachSrcNextTag((void *) (uCur+0x10));
		pDmaWrap->EndPacket();
	}
}

extern "C" void VU_vu1draw_start(void);
extern "C" void VU_vu1draw_end(void);

// distribute sps2 specific global pointers to relevant classes
void  DistribGlobPtrs(void * pDescCached, void * pDescUncached)
{
	CDmaWrap::m_pDescCached = (sps2Memory_t *) pDescCached;
	CDmaWrap::m_pDescUncached = (sps2Memory_t *) pDescUncached;
}
int iSPS2Device;

void delete_screen()
{
	if (cScreen!=NULL) delete cScreen;
	cScreen=NULL;
	printf("yay\n");
}

int main()
{
	// Open sps2-library and PS2Linux stuff
	iSPS2Device = sps2Init();

	if (iSPS2Device<0) {
	  printf("Can't open SPS2 device.  Is the module loaded?\n");
	  exit(-1);
	}

	// prealloc memory used by the application
	uint32 uAllocDmaMegs = 4;
	sps2Memory_t * pDescCached = sps2Allocate(uAllocDmaMegs*1024*1024, SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED, iSPS2Device);
	sps2Memory_t * pDescUncached = sps2Remap(pDescCached, SPS2_MAP_UNCACHED, iSPS2Device);
	DistribGlobPtrs(pDescCached, pDescUncached);
	void * pAllDmaMem = pDescCached->pvStart;
	int iVPU1; iVPU1=open("/dev/ps2vpu1", O_RDWR);
	for (int sig=0; sig<128; sig++)	// catch as many signals as possible
		signal(sig, sig_handle);

	// setup scene
	CSceneCreatorIntmd scene;
	scene.LoadToScene("scene_clones.asc", "geomdata_clones.bin");

	void * pMeshData = (void *) pAllDmaMem;
	CDmaVif1 cDmaVif1(pMeshData, (uAllocDmaMegs*1024*1024)/16);
	C3dsMesh meshtmp;
	meshtmp.SetupMesh(&cDmaVif1, &scene.GetSceneIntmd().GetMesh(0));
	myMesh = meshtmp;

	// cleanup unused data
	scene.CleanUp();
	
	// create application instances
	// intmdloader should really dictate
	// the number of instances
	uint32 uEnd = (uint32) pDescUncached->pvEnd;
	uint32 uOverFlowSize = uEnd - ((uint32) cDmaVif1.GetCurLocation());
	void * pCurLoc = cDmaVif1.GetCurLocation();

	for(int iNum=0; iNum<NUM_INSTANCES; iNum++)
	{
		CDmaVif1 cDmaVif1Inst(pCurLoc, uOverFlowSize>>4);
		pMeshInst[iNum] = myMesh.CreateInstance(&cDmaVif1Inst);
		pCurLoc = pMeshInst[iNum]->GetDmaChain()->GetCurLocation();
		uOverFlowSize = uEnd - ((uint32) pMeshInst[iNum]->GetDmaChain()->GetCurLocation());
	}
	
	// Upload VU1Code
	uint32 iLengthCodeInQuads = (((uint32)&VU_vu1draw_end - (uint32)&VU_vu1draw_start) + 15) >> 4;
	void * pDst = (void *) VU1_MICRO_MEM;
	void * pSrc = (void *) &VU_vu1draw_start;
	memcpy(pDst, pSrc, iLengthCodeInQuads*16);
	
	// Setup Screen
	cScreen = new CScreen;
	atexit(delete_screen);
	cScreen->ScreenSetup();	
	
	// Setup a Camera
	float fScrWdth = cScreen->GetWidth();
	float fScrHgt = cScreen->GetHeight();
	float fDepthZ = cScreen->GetDepth();
	CCamera cCam;
	cCam.SetBitmapSize(fScrWdth, -fScrHgt, fDepthZ);
	cCam.SetViewPortOffset(2048, 2048, fDepthZ);
	cCam.SetClipPlanes(-10, -10000);
	cCam.SetEyePos(Vec3(0,0,1));
	float fFov = 0.524f;	// total view of 60 degrees
	float fMinU = -tanf(fFov);
	float fMinV = (fMinU*3)/4;
	cCam.SetBoundVRC(fMinU, fMinV, -fMinU, -fMinV);
	cCam.SetPosViewPlane(Vec3(0,0,600));
	cCam.SetRotViewPlane(Vec3(0,0,1), Vec3(0,1,0));	

	// prelaunch some vif regs
	const uint32 uVifCodeIni[] _ALIGN16 =
	{
		OFFSET((1024-78)>>1),
		BASE(0),
		MSCALL(4),
		NOP
	};
	DPUT_EE_VIF1_FIFO(((uint128 *) uVifCodeIni)[0]);
	
	// setup call chain
	CDmaVif1 cRenderChain(pCurLoc, uOverFlowSize>>4);
	//cRenderChain.DisableTTE();
		
	// Flush Cache
	sps2FlushCache(iSPS2Device);

	float fAx=0, fAy=0, fAz=0;

	while(loop_again)
	{
		// angles
		const float PI = (3.141592f);
		const float delta_rot = 0.5f * PI / 180.0f;
		const float fXLength = 400;
		fAx += delta_rot     ; if(fAx > PI) fAx -= 2.0f * PI;
		fAy += delta_rot/2.0f; if(fAy > PI) fAy -= 2.0f * PI;
		fAz += delta_rot/3.0f; if(fAz > PI) fAz -= 2.0f * PI;

		// Reset render chain
		cRenderChain.Reset();

		for(int iNum=0; iNum<NUM_INSTANCES; iNum++)
		{
			if((iNum&0x1)!=0)
				fAx = -fAx;
			if((iNum&0x2)!=0)
				fAy = -fAy;
			if((iNum&0x4)!=0)
				fAz = -fAz;

			Mat44 mTrans;
			LoadRotation(&mTrans, fAx, fAy, fAz);
			//SetColumn(&mTrans, 3, Vec3(0,0,0));
			SetColumn(&mTrans, 3, Vec4(-fXLength + 2*(float(iNum)/(NUM_INSTANCES-1))*fXLength, 0, -500, 1) );


			// This just delivers the transformation for upload
			pMeshInst[iNum]->Draw(mTrans, cCam.GetWorldToBitmapTransform());

			if((iNum&0x1)!=0)
				fAx = -fAx;
			if((iNum&0x2)!=0)
				fAy = -fAy;
			if((iNum&0x4)!=0)
				fAz = -fAz;
	
			// Adding calltags (subdmachains) to global dmachain
			SyncStitch(&cRenderChain);
			cRenderChain.AttachSrcCallTag(pMeshInst[iNum]->GetDMA_Tags());
			cRenderChain.EndPacket();	
			SyncStitch(&cRenderChain);
			cRenderChain.AttachSrcCallTag(myMesh.GetDMA_Tags());	// all geometry
			cRenderChain.EndPacket();
		}

		// closeup global dmachain
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcEndTag();
		cRenderChain.EndPacket();
		cRenderChain.Launch(iSPS2Device);

		// spin locking
		cRenderChain.Wait();
		
		cScreen->Swap();
	}

	return 0;
}
